import json
from six import binary_type, text_type
from testtools import TestCase
from testtools.matchers import Equals, IsInstance

from eliottree._compat import dump_json_bytes


class DumpJsonBytesTests(TestCase):
    """
    Tests for ``eliottree._compat.dump_json_bytes``.
    """
    def test_text(self):
        """
        Text results are encoded as UTF-8.
        """
        def dump_text(obj):
            result = json.dumps(obj)
            if isinstance(result, binary_type):
                return result.decode('utf-8')
            return result

        result = dump_json_bytes(
            {'a': 42}, dumps=dump_text)
        self.assertThat(
            result,
            IsInstance(binary_type))
        self.assertThat(
            result,
            Equals(b'{"a": 42}'))

    def test_binary(self):
        """
        Binary results are left as-is.
        """
        def dump_binary(obj):
            result = json.dumps(obj)
            if isinstance(result, text_type):
                return result.encode('utf-8')
            return result

        result = dump_json_bytes(
            {'a': 42}, dumps=dump_binary)
        self.assertThat(
            result,
            IsInstance(binary_type))
        self.assertThat(
            result,
            Equals(b'{"a": 42}'))



from datetime import datetime

import jmespath
from iso8601.iso8601 import UTC


def filter_by_jmespath(query):
    """
    Produce a function for filtering a task by a jmespath query expression.
    """
    def _filter(task):
        return bool(expn.search(task))
    expn = jmespath.compile(query)
    return _filter


def filter_by_uuid(task_uuid):
    """
    Produce a function for filtering tasks by their UUID.
    """
    return filter_by_jmespath(u'task_uuid == `{}`'.format(task_uuid))


def _parse_timestamp(timestamp):
    """
    Parse a timestamp into a UTC L{datetime}.
    """
    return datetime.utcfromtimestamp(timestamp).replace(tzinfo=UTC)


def filter_by_start_date(start_date):
    """
    Produce a function for filtering by task timestamps after (or on) a certain
    date and time.
    """
    def _filter(task):
        return _parse_timestamp(task[u'timestamp']) >= start_date
    return _filter


def filter_by_end_date(end_date):
    """
    Produce a function for filtering by task timestamps before a certain date
    and time.
    """
    def _filter(task):
        return _parse_timestamp(task[u'timestamp']) < end_date
    return _filter


__all__ = [
    'filter_by_jmespath', 'filter_by_uuid', 'filter_by_start_date',
    'filter_by_end_date',
]



from calendar import timegm
from datetime import datetime

from iso8601.iso8601 import UTC
from testtools import TestCase
from testtools.matchers import Equals

from eliottree import (
    filter_by_end_date, filter_by_jmespath, filter_by_start_date,
    filter_by_uuid)
from eliottree.test.tasks import action_task, message_task


class FilterByJmespath(TestCase):
    """
    Tests for ``eliottree.filter_by_jmespath``.
    """
    def test_no_match(self):
        """
        Return ``False`` if the jmespath does not match the input.
        """
        self.assertThat(
            filter_by_jmespath('action_type == `app:action`')(message_task),
            Equals(False))

    def test_match(self):
        """
        Return ``True`` if the jmespath does match the input.
        """
        self.assertThat(
            filter_by_jmespath('action_type == `app:action`')(action_task),
            Equals(True))


class FilterByUUID(TestCase):
    """
    Tests for ``eliottree.filter_by_uuid``.
    """
    def test_no_match(self):
        """
        Return ``False`` if the input is not the specified task UUID.
        """
        self.assertThat(
            filter_by_uuid('nope')(message_task),
            Equals(False))

    def test_match(self):
        """
        Return ``True`` if the input is the specified task UUID.
        """
        self.assertThat(
            filter_by_uuid('cdeb220d-7605-4d5f-8341-1a170222e308')(
                message_task),
            Equals(True))


class FilterByStartDate(TestCase):
    """
    Tests for ``eliottree.filter_by_start_date``.
    """
    def test_no_match(self):
        """
        Return ``False`` if the input task's timestamp is before the start
        date.
        """
        now = datetime(2015, 10, 30, 22, 1, 15).replace(tzinfo=UTC)
        task = dict(
            message_task,
            timestamp=timegm(datetime(2015, 10, 30, 22, 1, 0).utctimetuple()))
        self.assertThat(
            filter_by_start_date(now)(task),
            Equals(False))

    def test_match(self):
        """
        Return ``True`` if the input task's timestamp is after the start date.
        """
        now = datetime(2015, 10, 30, 22, 1, 15).replace(tzinfo=UTC)
        task = dict(
            message_task,
            timestamp=timegm(datetime(2015, 10, 30, 22, 2).utctimetuple()))
        self.assertThat(
            filter_by_start_date(now)(task),
            Equals(True))

    def test_match_moment(self):
        """
        Return ``True`` if the input task's timestamp is the same as the start
        date.
        """
        now = datetime(2015, 10, 30, 22, 1, 15).replace(tzinfo=UTC)
        task = dict(
            message_task,
            timestamp=timegm(datetime(2015, 10, 30, 22, 1, 15).utctimetuple()))
        self.assertThat(
            filter_by_start_date(now)(task),
            Equals(True))


class FilterByEndDate(TestCase):
    """
    Tests for ``eliottree.filter_by_end_date``.
    """
    def test_no_match(self):
        """
        Return ``False`` if the input task's timestamp is after the start
        date.
        """
        now = datetime(2015, 10, 30, 22, 1, 15).replace(tzinfo=UTC)
        task = dict(
            message_task,
            timestamp=timegm(datetime(2015, 10, 30, 22, 2).utctimetuple()))
        self.assertThat(
            filter_by_end_date(now)(task),
            Equals(False))

    def test_no_match_moment(self):
        """
        Return ``False`` if the input task's timestamp is the same as the start
        date.
        """
        now = datetime(2015, 10, 30, 22, 1, 15).replace(tzinfo=UTC)
        task = dict(
            message_task,
            timestamp=timegm(datetime(2015, 10, 30, 22, 1, 15).utctimetuple()))
        self.assertThat(
            filter_by_end_date(now)(task),
            Equals(False))

    def test_match(self):
        """
        Return ``True`` if the input task's timestamp is before the start date.
        """
        now = datetime(2015, 10, 30, 22, 1, 15).replace(tzinfo=UTC)
        task = dict(
            message_task,
            timestamp=timegm(datetime(2015, 10, 30, 22, 1).utctimetuple()))
        self.assertThat(
            filter_by_end_date(now)(task),
            Equals(True))



"""
Tests for the command-line itself.
"""
from pprint import pformat
from subprocess import STDOUT, CalledProcessError, check_output
from tempfile import NamedTemporaryFile
from unittest import TestCase

from eliottree._compat import dump_json_bytes
from eliottree.test.tasks import message_task, missing_uuid_task


rendered_message_task = (
    u'cdeb220d-7605-4d5f-8341-1a170222e308\n'
    u'\u2514\u2500\u2500 twisted:log/1\n'
    u'    \u251c\u2500\u2500 eliot/timestamp: 2015-03-03 04:25:00\n'
    u'    \u251c\u2500\u2500 error: False\n'
    u'    \u2514\u2500\u2500 message: Main loop terminated.\n\n'
).encode('utf-8')


class EndToEndTests(TestCase):
    """
    Tests that actually run the command-line tool.
    """
    def test_stdin(self):
        """
        ``eliot-tree`` can read and render JSON messages from stdin when no
        arguments are given.
        """
        f = NamedTemporaryFile()
        f.write(dump_json_bytes(message_task))
        f.flush()
        f.seek(0, 0)
        self.assertEqual(check_output(["eliot-tree"], stdin=f),
                         rendered_message_task)

    def test_file(self):
        """
        ``eliot-tree`` can read and render JSON messages from a file on the
        command line.
        """
        f = NamedTemporaryFile()
        f.write(dump_json_bytes(message_task))
        f.flush()
        self.assertEqual(check_output(["eliot-tree", f.name]),
                         rendered_message_task)

    def test_json_parse_error(self):
        """
        ``eliot-tree`` displays an error containing the file name, line number
        and offending line in the event that JSON parsing fails.
        """
        f = NamedTemporaryFile()
        f.write(dump_json_bytes(message_task) + b'\n')
        f.write(b'totally not valid JSON {')
        f.flush()
        with self.assertRaises(CalledProcessError) as m:
            check_output(['eliot-tree', f.name], stderr=STDOUT)
        lines = m.exception.output.splitlines()
        first_line = lines[0].decode('utf-8')
        second_line = lines[1].decode('utf-8')
        self.assertIn('JSON parse error', first_line)
        self.assertIn(f.name, first_line)
        self.assertIn('line 2', first_line)
        self.assertEqual('totally not valid JSON {', second_line)

    def test_eliot_parse_error(self):
        """
        ``eliot-tree`` displays an error containing the original file name,
        line number and offending task in the event that parsing the message
        dict fails.
        """
        f = NamedTemporaryFile()
        f.write(dump_json_bytes(missing_uuid_task) + b'\n')
        f.flush()
        with self.assertRaises(CalledProcessError) as m:
            check_output(['eliot-tree', f.name], stderr=STDOUT)
        lines = m.exception.output.splitlines()
        first_line = lines[0].decode('utf-8')
        self.assertIn('Eliot message parse error', first_line)
        self.assertIn(f.name, first_line)
        self.assertIn('line 1', first_line)
        formatted = pformat(missing_uuid_task).encode('utf-8').splitlines()
        self.assertEqual(
            lines[1:len(formatted) + 1],
            formatted)



import sys
import traceback
from functools import partial

from eliot._action import WrittenAction
from eliot._message import WrittenMessage
from eliot._parse import Task
from six import text_type
from termcolor import colored
from toolz import compose, excepts, identity
from tree_format import format_tree

from eliottree import format
from eliottree._util import eliot_ns, format_namespace, is_namespace


RIGHT_DOUBLE_ARROW = u'\u21d2'

DEFAULT_IGNORED_KEYS = set([
    u'action_status', u'action_type', u'task_level', u'task_uuid',
    u'message_type'])


class Color(object):
    def __init__(self, color, attrs=[]):
        self.color = color
        self.attrs = attrs

    def __get__(self, instance, owner):
        return lambda text: instance.colored(
            text, self.color, attrs=self.attrs)


class COLORS(object):
    root = Color('white', ['bold'])
    parent = Color('magenta')
    success = Color('green')
    failure = Color('red')
    prop = Color('blue')
    error = Color('red', ['bold'])

    def __init__(self, colored):
        self.colored = colored


def _no_color(text, *a, **kw):
    """
    Colorizer that does not colorize.
    """
    return text


def _default_value_formatter(human_readable, field_limit, encoding='utf-8'):
    """
    Create a value formatter based on several user-specified options.
    """
    fields = {}
    if human_readable:
        fields = {
            eliot_ns(u'timestamp'): format.timestamp(),
        }
    return compose(
        # We want tree-format to handle newlines.
        partial(format.escape_control_characters, overrides={0x0a: u'\n'}),
        partial(format.truncate_value,
                field_limit) if field_limit else identity,
        format.some(
            format.fields(fields),
            format.text(),
            format.binary(encoding),
            format.anything(encoding)))


def message_name(colors, message, end_message=None):
    """
    Derive the name for a message.

    If the message is an action type then the ``action_type`` field is used in
    conjunction with ``task_level`` and ``action_status``. If the message is a
    message type then the ``message_type`` and ``task_level`` fields are used,
    otherwise no name will be derived.
    """
    if message is not None:
        if u'action_type' in message.contents:
            action_type = format.escape_control_characters(
                message.contents.action_type)
            if end_message:
                action_status = end_message.contents.action_status
            else:
                action_status = message.contents.action_status
            status_color = identity
            if action_status == u'succeeded':
                status_color = colors.success
            elif action_status == u'failed':
                status_color = colors.failure
            return u'{}{} {} {}'.format(
                colors.parent(action_type),
                message.task_level.to_string(),
                RIGHT_DOUBLE_ARROW,
                status_color(message.contents.action_status))
        elif u'message_type' in message.contents:
            message_type = format.escape_control_characters(
                message.contents.message_type)
            return u'{}{}'.format(
                colors.parent(message_type),
                message.task_level.to_string())
    return u'<unnamed>'


def format_node(format_value, colors, node):
    """
    Format a node for display purposes.

    Different representations exist for the various types of node:
        - `eliot._parse.Task`: A task UUID.
        - `eliot._action.WrittenAction`: An action's type, level and status.
        - `eliot._message.WrittenMessage`: A message's type and level.
        - ``tuple``: A field name and value.
    """
    if isinstance(node, Task):
        return u'{}'.format(
            colors.root(
                format.escape_control_characters(node.root().task_uuid)))
    elif isinstance(node, WrittenAction):
        return message_name(colors, node.start_message, node.end_message)
    elif isinstance(node, WrittenMessage):
        return message_name(colors, node)
    elif isinstance(node, tuple):
        key, value = node
        if isinstance(value, (dict, list)):
            value = u''
        else:
            value = format_value(value, key)
        if is_namespace(key):
            key = format_namespace(key)
        return u'{}: {}'.format(
            colors.prop(format.escape_control_characters(key)),
            value)
    raise NotImplementedError()


def message_fields(message, ignored_fields):
    """
    Sorted fields for a `WrittenMessage`.
    """
    def _items():
        try:
            yield eliot_ns('timestamp'), message.timestamp
        except KeyError:
            pass
        for key, value in message.contents.items():
            if key not in ignored_fields:
                yield key, value

    def _sortkey(x):
        k = x[0]
        return format_namespace(k) if is_namespace(k) else k
    return sorted(_items(), key=_sortkey) if message else []


def get_children(ignored_fields, node):
    """
    Retrieve the child nodes for a node.

    The various types of node have different concepts of children:
        - `eliot._parse.Task`: The root ``WrittenAction``.
        - `eliot._action.WrittenAction`: The start message fields, child
          ``WrittenAction`` or ``WrittenMessage``s, and end ``WrittenMessage``.
        - `eliot._message.WrittenMessage`: Message fields.
        - ``tuple``: Contained values for `dict` and `list` types.
    """
    if isinstance(node, Task):
        return [node.root()]
    elif isinstance(node, WrittenAction):
        return filter(None,
                      message_fields(node.start_message, ignored_fields) +
                      list(node.children) +
                      [node.end_message])
    elif isinstance(node, WrittenMessage):
        return message_fields(node, ignored_fields)
    elif isinstance(node, tuple):
        value = node[1]
        if isinstance(value, dict):
            return sorted(value.items())
        elif isinstance(value, list):
            return enumerate(value)
    return []


def track_exceptions(f, caught, default=None):
    """
    Decorate ``f`` with a function that traps exceptions and appends them to
    ``caught``, returning ``default`` in their place.
    """
    def _catch(_):
        caught.append(sys.exc_info())
        return default
    return excepts(Exception, f, _catch)


def render_tasks(write, tasks, field_limit=0, ignored_fields=None,
                 human_readable=False, colorize=False, write_err=None,
                 format_node=format_node, format_value=None):
    """
    Render Eliot tasks as an ASCII tree.

    :type write: ``Callable[[text_type], None]``
    :param write: Callable used to write the output.
    :type tasks: ``Iterable``
    :param tasks: Iterable of parsed Eliot tasks, as returned by
    `eliottree.tasks_from_iterable`.
    :param int field_limit: Length at which to begin truncating, ``0`` means no
    truncation.
    :type ignored_fields: ``Set[text_type]``
    :param ignored_fields: Set of field names to ignore, defaults to ignoring
    most Eliot metadata.
    :param bool human_readable: Render field values as human-readable?
    :param bool colorize: Colorized the output?
    :type write_err: Callable[[`text_type`], None]
    :param write_err: Callable used to write errors.
    :param format_node: See `format_node`.
    :type format_value: Callable[[Any], `text_type`]
    :param format_value: Callable to format a value.
    """
    if ignored_fields is None:
        ignored_fields = DEFAULT_IGNORED_KEYS
    colors = COLORS(colored if colorize else _no_color)
    caught_exceptions = []
    if format_value is None:
        format_value = _default_value_formatter(
            human_readable=human_readable,
            field_limit=field_limit)
    _format_value = track_exceptions(
        format_value,
        caught_exceptions,
        u'<value formatting exception>')
    _format_node = track_exceptions(
        partial(format_node, _format_value, colors),
        caught_exceptions,
        u'<node formatting exception>')
    _get_children = partial(get_children, ignored_fields)
    for task in tasks:
        write(format_tree(task, _format_node, _get_children))
        write(u'\n')

    if write_err and caught_exceptions:
        write_err(
            colors.error(
                u'Exceptions ({}) occurred during processing:\n'.format(
                    len(caught_exceptions))))
        for exc in caught_exceptions:
            for line in traceback.format_exception(*exc):
                if not isinstance(line, text_type):
                    line = line.decode('utf-8')
                write_err(line)
            write_err(u'\n')


__all__ = ['render_tasks']



from functools import wraps
from warnings import warn

from six import text_type
from termcolor import colored
from tree_format import format_tree

from eliottree._render import (
    COLORS, DEFAULT_IGNORED_KEYS, _default_value_formatter, _no_color)
from eliottree._util import is_namespace, eliot_ns
from eliottree.format import escape_control_characters


def render_task_nodes_unicode(write, nodes, field_limit,
                              ignored_task_keys=None, human_readable=False,
                              colorize=False):
    """
    Render a tree of task nodes as an ``ASCII`` tree.

    :type write: ``callable`` taking a single ``text_type`` argument
    :param write: Callable to write the output.

    :type nodes: ``list`` of ``(text_type, _TaskNode)``.
    :param nodes: List of pairs of task UUID and task nodes, as obtained
        from ``Tree.nodes``.

    :type field_limit: ``int``
    :param field_limit: Length at which to begin truncating, ``0`` means no
        truncation.

    :type ignored_task_keys: ``set`` of ``text_type``
    :param ignored_task_keys: Set of task key names to ignore.

    :type human_readable: ``bool``
    :param human_readable: Should this be rendered as human-readable?

    :type colorize: ``bool``
    :param colorize: Should the output be colorized?
    """
    if ignored_task_keys is None:
        ignored_task_keys = DEFAULT_IGNORED_KEYS
    colors = COLORS(colored if colorize else _no_color)
    format_value = _default_value_formatter(
        human_readable=human_readable,
        field_limit=field_limit)
    get_children = get_children_factory(
        ignored_task_keys,
        format_value)
    get_name = get_name_factory(colors)
    for root in nodes:
        write(format_tree(root, get_name, get_children))
        write(u'\n')


def render_task_nodes(write, nodes, field_limit,
                      ignored_task_keys=None, human_readable=False,
                      encoding='utf-8', colorize=False):
    """
    :type write: ``callable`` taking a single ``bytes_type`` argument
    :param write: Callable to write the output.

    :param str encoding: Encoding for data to be written to ``write``.

    :seealso: `render_task_nodes`
    """
    warn('render_task_nodes is deprecated, use eliottree.render_tasks instead',
         DeprecationWarning, 2)
    render_task_nodes_unicode(
        write=lambda value: write(value.encode(encoding)),
        nodes=nodes,
        field_limit=field_limit,
        ignored_task_keys=ignored_task_keys,
        human_readable=human_readable,
        colorize=colorize)


def get_name_factory(colors):
    """
    Create a ``get_name`` function for use with `format_tree`.
    """
    def get_name(task):
        if isinstance(task, text_type):
            return escape_control_characters(task)
        elif isinstance(task, tuple):
            key = task[0]
            if is_namespace(key):
                key = key.name
            name = escape_control_characters(key)
            if isinstance(task[1], dict):
                return name
            elif isinstance(task[1], text_type):
                return u'{}: {}'.format(
                    colors.prop(name),
                    # No need to escape this because we assume the value
                    # formatter did that already.
                    task[1])
            else:
                return colors.root(name)
        else:
            name = escape_control_characters(task.name)
            if task.success is True:
                return colors.success(name)
            elif task.success is False:
                return colors.failure(name)
            return name
    return get_name


def listify(f):
    """
    Convert the decorated function's result to a ``list``.
    """
    @wraps(f)
    def _wrapper(*a, **kw):
        return list(f(*a, **kw))
    return _wrapper


def get_children_factory(ignored_task_keys, format_value):
    """
    Create a ``get_children`` function for use with `format_tree`.

    :type field_limit: ``int``
    :param field_limit: Length at which to begin truncating, ``0`` means no
        truncation.

    :type ignored_task_keys: ``set`` of ``text_type``
    :param ignored_task_keys: Set of task key names to ignore.

    :type format_value: ``Callable[[Any, text_type], Any]``
    :param format_value: Function to format a value for display purposes.
    """
    def items_children(items):
        for key, value in sorted(items):
            if key not in ignored_task_keys:
                if key == u'timestamp':
                    key = eliot_ns(key)
                if isinstance(value, dict):
                    yield key, value
                else:
                    yield key, format_value(value, key)

    @listify
    def get_children(task):
        if isinstance(task, text_type):
            return
        elif isinstance(task, tuple):
            if isinstance(task[1], dict):
                for child in items_children(task[1].items()):
                    yield child
            elif isinstance(task[1], text_type):
                # The value of Unicode values is incorporated into the name.
                return
            else:
                yield task[1]
            return
        else:
            for child in items_children(task.task.items()):
                if child[0] == u'timestamp':
                    yield eliot_ns(child[0]), child[1]
                else:
                    yield child
            for child in task.children():
                yield child
    return get_children


__all__ = ['render_task_nodes']



from testtools import TestCase
from testtools.matchers import (
    Contains, Equals, Is, IsDeprecated, MatchesListwise, raises)

from eliottree.test.tasks import (
    action_task, action_task_end, message_task, nested_action_task,
    unnamed_message)
from eliottree.tree import Tree, _TaskNode, missing_start_task, task_name


def _flattened_tasks(nodes):
    """
    Construct a flat iterable of all the tasks for an iterable of nodes.
    """
    for n in nodes:
        yield n.task
        for c in n.children():
            yield c.task


class TaskNameTests(TestCase):
    """
    Tests for ``eliottree.tree.task_name``.
    """
    def test_none(self):
        """
        Cannot create a task name for ``None``.
        """
        self.assertThat(
            lambda: task_name(None),
            raises(ValueError))

    def test_message_type(self):
        """
        If the task includes a ``message_type`` key use it to construct the
        name.
        """
        self.assertThat(
            task_name(message_task),
            Equals(u'twisted:log@1'))

    def test_no_message_type(self):
        """
        If the task does not include a ``message_type`` key use the
        ``action_type`` and ``action_status`` keys to construct a name.
        """
        self.assertThat(
            task_name(action_task),
            Equals(u'app:action@1/started'))

    def test_no_action_type(self):
        """
        If the task does not include either a ``message_type`` or an
        ``action_type`` key, then return None.
        """
        self.assertThat(
            task_name(unnamed_message),
            Is(None))

    def test_levels(self):
        """
        Include the task level in the task name.
        """
        self.assertThat(
            task_name(nested_action_task),
            Equals(u'app:action:nest@1,1/started'))


class TaskNodeTests(TestCase):
    """
    Tests for ``eliottree.tree._TaskNode``.
    """
    def test_none(self):
        """
        Passing ``None`` as the task value raises ``ValueError``
        """
        self.assertThat(
            lambda: _TaskNode(task=None),
            raises(ValueError))

    def test_repr(self):
        """
        Representation of a task node.
        """
        node = _TaskNode(task=action_task)
        self.assertThat(
            repr(node),
            Equals("<_TaskNode 'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4' "
                   "'app:action@1/started' children=0>"))

    def test_repr_nameless(self):
        """
        Representation of a task without a name.
        """
        node = _TaskNode(task=unnamed_message)
        self.assertThat(
            repr(node),
            Equals("<_TaskNode 'cdeb220d-7605-4d5f-8341-1a170222e308' "
                   "'<UNNAMED TASK>' children=0>"))

    def test_repr_childen(self):
        """
        Representation of a task node with children.
        """
        node = _TaskNode(task=action_task, name=u'foo')
        node.add_child(_TaskNode(task=message_task))
        self.assertThat(
            repr(node),
            Equals("<_TaskNode 'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4' "
                   "'foo' children=1>"))

    def test_no_children(self):
        """
        ``_TaskNode.children`` returns an empty list for a node with no
        children.
        """
        node = _TaskNode(task=action_task, name=u'foo')
        self.assertThat(
            node.children(),
            Equals([]))

    def test_children(self):
        """
        ``_TaskNode.children`` returns an list of child nodes sorted by their
        level regardless of the order they were added.
        """
        node = _TaskNode(task=action_task, name=u'foo')
        child = _TaskNode(task=nested_action_task)
        child2 = _TaskNode(task=action_task_end)
        node.add_child(child2)
        node.add_child(child)
        self.assertThat(
            node.children(),
            Equals([child, child2]))

    def test_nested_children(self):
        """
        ``_TaskNode.children`` does not include grandchildren.
        """
        node = _TaskNode(task=action_task, name=u'foo')
        child = _TaskNode(task=message_task)
        node.add_child(child)
        child2 = _TaskNode(task=nested_action_task)
        node.add_child(child2)
        self.assertThat(
            node.children(),
            Equals([child]))
        self.assertThat(
            child.children(),
            Equals([child2]))


class TreeTests(TestCase):
    """
    Tests for ``eliottree.tree.Tree``.
    """
    def test_deprecated(self):
        """
        `Tree` is deprecated.
        """
        self.assertThat(
            Tree,
            IsDeprecated(Contains('Tree is deprecated')))

    def test_initial(self):
        """
        The initial state of a tree is always empty.
        """
        tree = Tree()
        self.assertThat(tree.nodes(), Equals([]))

    def test_merge_tasks(self):
        """
        Merge tasks into the tree and retrieve an list of key-node
        pairs ordered by task timestamp.
        """
        tree = Tree()
        matches = tree.merge_tasks([message_task, action_task])
        self.expectThat(matches, Is(None))
        keys, nodes = zip(*tree.nodes())
        self.expectThat(
            list(keys),
            Equals(['cdeb220d-7605-4d5f-8341-1a170222e308',
                    'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4']))
        self.assertThat(
            list(_flattened_tasks(nodes)),
            MatchesListwise([Equals(message_task),
                             Equals(action_task)]))

    def test_merge_nested_tasks(self):
        """
        Merge nested tasks into the tree and retrieve an list of key-node
        pairs ordered by task timestamp.
        """
        tree = Tree()
        matches = tree.merge_tasks([action_task_end, action_task])
        self.expectThat(matches, Is(None))
        keys, nodes = zip(*tree.nodes())
        self.expectThat(
            list(keys),
            Equals(['f3a32bb3-ea6b-457c-aa99-08a3d0491ab4']))
        self.assertThat(
            list(_flattened_tasks(nodes)),
            MatchesListwise([Equals(action_task),
                             Equals(action_task_end)]))

    def test_merge_startless_tasks(self):
        """
        Merging a task that will never have a start parent creates a fake start
        task.
        """
        tree = Tree()
        missing_task = missing_start_task(action_task_end)
        matches = tree.merge_tasks([action_task_end])
        self.expectThat(matches, Is(None))
        keys, nodes = zip(*tree.nodes())
        self.expectThat(
            list(keys),
            Equals(['f3a32bb3-ea6b-457c-aa99-08a3d0491ab4']))
        self.assertThat(
            list(_flattened_tasks(nodes)),
            MatchesListwise([Equals(missing_task),
                             Equals(action_task_end)]))

    def test_merge_tasks_filtered(self):
        """
        Merge tasks into the tree with a filter function, generating a set of
        matches that can be used to prune the tree.
        """
        tree = Tree()
        filters = [lambda task: task.get(u'action_type') == u'app:action']
        matches = tree.merge_tasks([action_task, message_task], filters)
        keys, nodes = zip(*tree.nodes(matches))
        self.expectThat(
            list(keys),
            Equals(['f3a32bb3-ea6b-457c-aa99-08a3d0491ab4']))
        self.expectThat(
            list(keys),
            Equals(list(matches)))

        self.assertThat(
            list(_flattened_tasks(nodes)),
            MatchesListwise([Equals(action_task)]))

    def test_merge_tasks_no_matches(self):
        """
        Merging tasks when no tasks match returns an empty set of tasks.
        """
        tree = Tree()
        matches = tree.merge_tasks(
            [action_task, message_task], [lambda task: False])
        self.expectThat(list(matches), Equals([]))



from eliottree._errors import EliotParseError, JSONParseError
from eliottree._parse import tasks_from_iterable
from eliottree._render import render_tasks
from eliottree.filter import (
    filter_by_end_date, filter_by_jmespath, filter_by_start_date,
    filter_by_uuid)
from eliottree.render import render_task_nodes
from eliottree.tree import Tree


__version__ = '17.1.0'


__all__ = [
    'Tree', 'render_task_nodes', 'render_task_nodes_unicode',
    'filter_by_jmespath', 'filter_by_uuid', 'filter_by_start_date',
    'filter_by_end_date', 'render_tasks', 'tasks_from_iterable',
    'EliotParseError', 'JSONParseError',
]

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import argparse
import codecs
import json
import sys
from pprint import pformat

import iso8601
from six import PY3, binary_type, reraise
from six.moves import filter
from toolz import compose

from eliottree import (
    EliotParseError, JSONParseError, filter_by_end_date, filter_by_jmespath,
    filter_by_start_date, filter_by_uuid, render_tasks, tasks_from_iterable)


def text_writer(fd):
    """
    File writer that accepts Unicode to write.
    """
    if PY3:
        return fd
    return codecs.getwriter('utf-8')(fd)


def text_reader(fd):
    """
    File reader that returns Unicode from reading.
    """
    if PY3:
        return fd
    return codecs.getreader('utf-8')(fd)


def parse_messages(files=None, select=None, task_uuid=None, start=None,
                   end=None):
    """
    Parse message dictionaries from inputs into Eliot tasks, filtering by any
    provided criteria.
    """
    def filter_funcs():
        if task_uuid is not None:
            yield filter_by_uuid(task_uuid)
        if start:
            yield filter_by_start_date(start)
        if end:
            yield filter_by_end_date(end)
        if select is not None:
            for query in select:
                yield filter_by_jmespath(query)

    def _parse(files, inventory):
        for file in files:
            file_name = getattr(file, 'name', '<unknown>')
            for line_number, line in enumerate(file, 1):
                try:
                    task = json.loads(line)
                    inventory[id(task)] = file_name, line_number
                    yield task
                except Exception:
                    raise JSONParseError(
                        file_name,
                        line_number,
                        line.rstrip(u'\n'),
                        sys.exc_info())

    if not files:
        files = [text_reader(sys.stdin)]
    inventory = {}
    return inventory, tasks_from_iterable(
        filter(compose(*filter_funcs()), _parse(files, inventory)))


def display_tasks(tasks, color, ignored_fields, field_limit, human_readable):
    """
    Render Eliot tasks, apply any command-line-specified behaviour and render
    the task trees to stdout.
    """
    write = text_writer(sys.stdout).write
    write_err = text_writer(sys.stderr).write
    if color == 'auto':
        colorize = sys.stdout.isatty()
    else:
        colorize = color == 'always'
    render_tasks(
        write=write,
        write_err=write_err,
        tasks=tasks,
        ignored_fields=set(ignored_fields) or None,
        field_limit=field_limit,
        human_readable=human_readable,
        colorize=colorize)


def _decode_command_line(value, encoding='utf-8'):
    """
    Decode a command-line argument.
    """
    if isinstance(value, binary_type):
        return value.decode(encoding)
    return value


def main():
    parser = argparse.ArgumentParser(
        description='Display an Eliot log as a tree of tasks.')
    parser.add_argument('files',
                        metavar='FILE',
                        nargs='*',
                        type=argparse.FileType('r'),
                        help='''Files to process. Omit to read from stdin.''')
    parser.add_argument('-u', '--task-uuid',
                        dest='task_uuid',
                        metavar='UUID',
                        type=_decode_command_line,
                        help='''Select a specific task by UUID''')
    parser.add_argument('-i', '--ignore-task-key',
                        action='append',
                        default=[],
                        dest='ignored_fields',
                        metavar='KEY',
                        type=_decode_command_line,
                        help='''Ignore a task key, use multiple times to ignore
                        multiple keys. Defaults to ignoring most Eliot standard
                        keys.''')
    parser.add_argument('--raw',
                        action='store_false',
                        dest='human_readable',
                        help='''Do not format some task values (such as
                        UTC timestamps) as human-readable''')
    parser.add_argument('--color',
                        default='auto',
                        choices=['always', 'auto', 'never'],
                        dest='color',
                        help='''Color the output. Defaults based on whether
                        the output is a TTY.''')
    parser.add_argument('-l', '--field-limit',
                        metavar='LENGTH',
                        type=int,
                        default=100,
                        dest='field_limit',
                        help='''Limit the length of field values to LENGTH or a
                        newline, whichever comes first. Use a length of 0 to
                        output the complete value.''')
    parser.add_argument('--select',
                        action='append',
                        metavar='QUERY',
                        dest='select',
                        type=_decode_command_line,
                        help='''Select tasks to be displayed based on a jmespath
                        query, can be specified multiple times to mimic logical
                        AND. If any child task is selected the entire top-level
                        task is selected. See <http://jmespath.org/>''')
    parser.add_argument('--start',
                        dest='start',
                        type=iso8601.parse_date,
                        help='''Select tasks whose timestamp occurs after (or
                        on) an ISO8601 date.''')
    parser.add_argument('--end',
                        dest='end',
                        type=iso8601.parse_date,
                        help='''Select tasks whose timestamp occurs before an
                        ISO8601 date.''')
    args = parser.parse_args()
    stderr = text_writer(sys.stderr)
    try:
        inventory, tasks = parse_messages(
            files=args.files,
            select=args.select,
            task_uuid=args.task_uuid,
            start=args.start,
            end=args.end)
        display_tasks(
            tasks=tasks,
            color=args.color,
            ignored_fields=args.ignored_fields,
            field_limit=args.field_limit,
            human_readable=args.human_readable)
    except JSONParseError as e:
        stderr.write('JSON parse error, file {}, line {}:\n{}\n\n'.format(
            e.file_name,
            e.line_number,
            e.line))
        reraise(*e.exc_info)
    except EliotParseError as e:
        file_name, line_number = inventory.get(
            id(e.message_dict), (u'<unknown>', u'<unknown>'))
        stderr.write(
            'Eliot message parse error, file {}, line {}:\n{}\n\n'.format(
                file_name,
                line_number,
                pformat(e.message_dict)))
        reraise(*e.exc_info)



class EliotParseError(RuntimeError):
    """
    An error occurred while parsing a particular Eliot message dictionary.
    """
    def __init__(self, message_dict, exc_info):
        self.message_dict = message_dict
        self.exc_info = exc_info
        RuntimeError.__init__(self)


class JSONParseError(RuntimeError):
    """
    An error occurred while parsing JSON text.
    """
    def __init__(self, file_name, line_number, line, exc_info):
        self.file_name = file_name
        self.line_number = line_number
        self.line = line
        self.exc_info = exc_info


__all__ = ['EliotParseError', 'JSONParseError']
